""" This module and its sub-modules have been adapted from or taken as-is from and with
the permission of the authors/maintainers of NASLib, at
https://github.com/automl/NASLib. """

import logging

_log = logging.getLogger(__name__)
